
/*

    File: scan.c

    Copyright (C) 2004, 2005, 2007  by Wolfgang Zekoll <wzk@quietsche-entchen.de>
 
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <wait.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>

#include "lib.h"
#include "ip-lib.h"
#include "smtp.h"


#define	IS_CLEAN		0
#define	IS_VIRUS		1
#define	IS_SPAM			2


tokentab_t dirmodetab[] = {
	{ "autosubdir",		DIR_AUTOSUBDIR },
	{ "flat",		DIR_PLAIN },
	{ "plain",		DIR_PLAIN },
	{ "",			0 }
    };

tokentab_t indexmodetab[] = {
	{ "info",		INDEX_INFO },
	{ "table",		INDEX_TABLE },
	{ "none",		INDEX_NONE },
	{ "",			0 }
    };


static char tmpdir[200] =	"/tmp";

static struct _tmpfile {
    char	filename[200];
    struct _tmpfile *next;
    } *lasttmp = NULL;

static int count =		0;
static int pid =		0;

static char prefix[200] =	"";


char *gettmpdir()
{
	return (tmpdir);
}

int settmpdir(char *dir)
{
	copy_string(tmpdir, dir, sizeof(tmpdir));
	return (0);
}


static int unlink_existing(char *filename)
{
	struct stat sbuf;

	if (stat(filename, &sbuf) == 0) {
		if (unlink(filename) != 0)
			return (1);
		}
	
	return (0);
}


static void cleanup(void)
{
	struct stat sbuf;
	struct _tmpfile *tmp;

	if (pid == 0  ||  pid != getpid())
		return;

	tmp = lasttmp;
	while (tmp != NULL) {
		if (stat(tmp->filename, &sbuf) == 0) {
			unlink(tmp->filename);
			}

		/*
		 * cleanup() is called on program termination
		 * so we do not need to free the memory.
		 */

		tmp = tmp->next;
		}

	return;
}

static int init_tmplist()
{
	char	*p;
	struct _tmpfile *tmp, *next;

	if (pid == 0) {
		pid = getpid();
		atexit(cleanup);

		if ((p = getenv("TMPPREFIX")) == NULL  ||  *p == 0)
			strcpy(prefix, "smtp-");
		else
			copy_string(prefix, p, sizeof(prefix));
		}
	else if (pid != getpid()) {

		/*
		 * we are in a child process - let's erase the parent's
		 * tempfile list and create our own.
		 */

		pid     = getpid();
		count   = 0;

		tmp = lasttmp;
		while (tmp != NULL) {
			next = tmp->next;
			free(tmp);

			tmp = next;
			}
			
		lasttmp = NULL;
		if (atexit(cleanup) != 0)
			printerror(1, "-ERR", "can't register cleaup()");
		}
		
	return (0);
}

static char *gettmpfile(char *filename, int size)
{
	int	fd;
	struct _tmpfile *tmp;
	static unsigned long now = 0;

	if (now == 0)
		now = time(NULL);

	init_tmplist();
	if ((tmp = malloc(sizeof(struct _tmpfile))) == NULL)
		printerror(1, "-ERR", "memory allocation error");

	snprintf (tmp->filename, sizeof(tmp->filename) - 2, "%s/%s%ld-%05d.%03d.tmp",
			tmpdir, prefix, now, pid, ++count);
	
	tmp->next = lasttmp;
	lasttmp = tmp;

	unlink_existing(tmp->filename);
	if ((fd = open(tmp->filename, O_CREAT | O_WRONLY | O_TRUNC, 0600)) < 0) {
		printerror(1, "-ERR", "can't open tmpfile: %s, error= %s", tmp->filename, strerror(errno));
		exit (1);
		}


	close (fd);
	if (filename != NULL)
		copy_string(filename, tmp->filename, size);

	return (tmp->filename);
}


int createdir(char *dir)
{
	struct stat sbuf;

	if (stat(dir, &sbuf) != 0) {
		if (mkdir(dir, 0700) != 0) {
			printerror(1, "-ERR", "can't create directory %s, error= %s",
					dir, strerror(errno));
			}
		else if (S_ISDIR(sbuf.st_mode) == 0)
			printerror(1, "-ERR", "not a directory: %s", dir);
		}

	return (0);
}

int quarantine(smtp_t *x, qdir_t *qd, char *spoolfile, int status, char *par, char *val)
{
	char	*p, *q, *basename, str[200], newname[200], infofile[200];
	FILE	*fp;

	writestatfile(x, "QUARANTINE");
	if ((basename = strrchr(x->spoolfile, '/')) == NULL)
		basename = x->spoolfile;
	else
		basename++;

	/* q = x->config->clamav.quarantine; */
	q = qd->dir;
	if (qd->dirmode == DIR_AUTOSUBDIR) {
		int	exists;
		unsigned long now;
		char	date[40];
		struct tm tm;
		struct stat sbuf;
		static char dirname[200];

		now = time(NULL);
		tm = *localtime(&now);
		strftime(date, sizeof(date) - 2, "%Y-%m-%d", &tm);
		snprintf (dirname, sizeof(dirname) - 2, "%s/%s", q, date);
		if ((exists = stat(dirname, &sbuf)) == 0  &&  S_ISDIR(sbuf.st_mode) != 0)
			q = dirname;
		else if (exists != 0  &&  mkdir(dirname, 0755) == 0)
			q = dirname;
		}

	snprintf (newname, sizeof(newname) - 6, "%s/%s", q, basename);
	if (rename(x->spoolfile, newname) != 0) {
		printerror(1, "-ERR", "can't quarantine file %s, error= %s",
				newname, strerror(errno));
		}

	if (qd->indexstyle == DIR_AUTOSUBDIR) {
		copy_string(infofile, newname, sizeof(infofile));
		if ((p = strrchr(infofile, '.')) == NULL)
			printerror(0, "-ERR", "infofile filename error: %s", infofile);
		else {
			strcpy(p, ".info");
			if ((fp = fopen(infofile, "w")) == NULL) {
				printerror(1, "-ERR", "can't open infofile: %s, error= %s",
						infofile, strerror(errno));
				}
			else {
				fprintf (fp, "client: %s\n", x->client);
				fprintf (fp, "server: %s:%u\n", x->s.name, x->s.port);
				fprintf (fp, "sender: %s\n", x->sender);

				p = x->rcpt.list;
				while (*(p = skip_ws(p)) != 0) {
					get_word(&p, str, sizeof(str));
					fprintf (fp, "rcpt: %s\n", str);
					}

				fprintf (fp, "subject: %s\n", x->subject);
				if (status == IS_VIRUS)
					fprintf (fp, "virus: %s\n", val);

				fprintf (fp, "filename: %s\n", newname);
				fprintf (fp, "size: %lu\n", x->size);

				if (status == IS_SPAM)
					fprintf (fp, "spamscore: %s\n", val);

				fprintf (fp, "\n");
				fclose (fp);
				}
			}
		}
	else if (qd->indexstyle == INDEX_TABLE) {
		int	c, i;
		char	filename[300];
		string_t s;
		FILE	*fp;

		snprintf(filename, sizeof(filename) - 2, "%s/.index.tab", q);
		if ((fp = fopen(filename, "a")) == NULL) {
			printerror(1, "-ERR", "can't open indexfile: filename= %s, error= %s",
					filename, strerror(errno));
			}

		clearstring(&s);
		addline(&s, x->rcpt.list, "");
		for (i=0; (c = s.string[i]) != 0; i++) {
			if (c == ' ')
				s.string[i] = ',';
			}
		
		fprintf (fp, "%s\t%s:%u\t%s\t%s\t%s\t%s\t%s\t%lu\t%s\n",
				x->client, x->s.name, x->s.port, x->sender,
				s.string, x->subject,
				status == IS_VIRUS? val: "-",
				newname, x->size,
				status == IS_SPAM? val: "-");
		fclose (fp);
		}

	printerror(0, "+INFO", "quarantined e-mail= %s, %s= %s",
			basename, par, val);
	return (0);
}


int getclamversion(smtp_t *x)
{
	char	*p;

	memset(&x->clamav.bio, 0, sizeof(x->clamav.bio));
	if ((x->clamav.bio.fd = openip(x->clamav.adress, x->clamav.port, NULL, 0)) < 0)
		printerror(1, "-PROXY", "can't get clamav version: error= %s", strerror(errno));

	p = "VERSION\n";
	write(x->clamav.bio.fd, p, strlen(p));
	if (readline_fd(x, &x->clamav.bio, x->clamav.version, sizeof(x->clamav.version), 0, 1) == NULL  ||
	    *x->clamav.version == 0) {
	    	printerror(1, "-PROXY", "invalid clamav version: response= %s", x->clamav.version);
		}

	close(x->clamav.bio.fd);
	return (0);
}

int doscanmail(smtp_t *x, char *resp, int size)
{
	int	rc, virulent, viruscount, spamscore;
	char	*p, status[80], virusname[200], spamlevel[80], line[4096];
	FILE	*fp;

	writestatfile(x, "RECEIVE");
	virulent   = IS_CLEAN;
	spamscore  = 0;
	*spamlevel = 0;

	x->lasttimer = time(NULL);
	if (*x->spoolfile == 0)
		gettmpfile(x->spoolfile, sizeof(x->spoolfile));

	setvar("MAIL", x->spoolfile);

	/*
	 * Read the file from the server ...
	 */

	if ((fp = fopen(x->spoolfile, "w")) == NULL)
		printerror(1, "-PROXY", "can't write spoolfile: filename= %s, error= %s", x->spoolfile, strerror(errno));

	cfputs(x, "354 send data");
	x->size = 0;
	while (readline_fd(x, &x->cbio, line, sizeof(line), 1, 0) != NULL) {
		if (*(p = line) == '.') {
			p++;
			if (*p == 0)
				break;
			}

		x->size += strlen(line) + 1;
		fprintf (fp, "%s\n", p);
		}


	fclose (fp);

	if (debug != 0)
		printerror(0, "*DEBUG", "email complete, lasttimer= %lu", x->lasttimer);


	/*
	 * ... connect to the local clamav server ...
	 */

	if (x->config->scanmail != 0) {
		writestatfile(x, "CLAMSCAN");

		viruscount = 0;
		memset(&x->clamav.bio, 0, sizeof(x->clamav.bio));
		copy_string(x->clamav.bio.name, "clamav", sizeof(x->clamav.bio.name));

		if (*x->clamav.adress == 0)
			strcpy(x->clamav.adress, "127.0.0.1:3310");

		x->clamav.port = get_port(x->clamav.adress, 3310);
		if ((x->clamav.bio.fd = openip(x->clamav.adress, x->clamav.port, NULL, 0)) < 0) {
			if (x->config->clamav.passonerror == 0)
				printerror(1, "-PROXY", "can't connect to clamav daemon: error= %s", strerror(errno));
			else {
				printerror(0, "-INFO", "can't connect to clamav daemon: error= %s", strerror(errno));
				viruscount = -1;
				copy_string(virusname, "clamav.connect", sizeof(virusname));
				goto passon;
				}
			}

		/*
		 * ... get the version ...
		 */

		/* getclamversion(x); */

		/*
		 * ... and scan the e-mail.
		 */

		snprintf (line, sizeof(line) - 2, "CONTSCAN %s\n", x->spoolfile);
		if (debug != 0)
			fprintf (stderr, ">>> CAV: %s\n", noctrl(line));

		write(x->clamav.bio.fd, line, strlen(line));
		while (1) {
			if (readline_fd(x, &x->clamav.bio, line, sizeof(line), 1, 1) == NULL)
				break;		/* Richtig oder falsch? */

			if (debug != 0)
				fprintf (stderr, "CAV >>>: %s\n", line);

			if ((p = strrchr(line, ' ')) == NULL) {
				if (x->config->clamav.passonerror == 0)
					printerror(1, "-PROXY", "clamav protocol error: response= %s", line);
				else {
					printerror(0, "-INFO", "clamav protocol error: response= %s", line);
					viruscount = -1;
					copy_string(virusname, "clamav.protocol", sizeof(virusname));
					break;
					}
				}

			p++;
			get_word(&p, status, sizeof(status));
			strupr(status);
			if (strcmp(status, "OK") == 0)
				break;
			else if (strcmp(status, "FOUND") == 0) {
				char	filename[200];

				if (viruscount == 1)
					printerror(1, "+INFO", "virus found: %s", virusname);

				p = line;
				get_quoted(&p, ':', filename, sizeof(filename));
				p = skip_ws(p);
				get_word(&p, virusname, sizeof(virusname));

				viruscount++;
				if (viruscount > 1)
					printerror(1, "+INFO", "virus found: %s", virusname);

				virulent = IS_VIRUS;
				}
			else {
				if (x->config->clamav.passonerror == 0)
					printerror(1, "-PROXY", "clamav protocol error: response= %s", line);
				else {
					printerror(0, "-INFO", "clamav protocol error: response= %s", line);
					viruscount = -1;
					copy_string(virusname, "clamav.response", sizeof(virusname));
					break;
					}
				}
			}

		close(x->clamav.bio.fd);
		x->lasttimer = 0;
		}

passon:
	if (virulent != 0) {
		writestatfile(x, "REJECT");
		
		/*
		 * Reject virulent e-mails.
		 */

		rc = 550;
		snprintf (resp, size - 2, "%3d virus detected, virusname= %s",
					x->config->clamav.statuscode, virusname);

		
		if (*x->config->clamav.qd.dir != 0) {
			quarantine(x, &x->config->clamav.qd, x->spoolfile, IS_VIRUS, "virusname", virusname);
			*x->spoolfile = 0;
			}

		return (rc);
		}


	if (viruscount > 0)
		printerror(0, "-INFO", "passing virulent e-mail count= %d, virusname= %s", viruscount, virusname);
	else if (viruscount < 0)
		printerror(0, "-INFO", "passing unscanned e-mail error= %s", virusname);
	else
		printerror(0, "+INFO", "virus clean");


	/*
	 * Emulate spam scan if required by RBL.  Configuration
	 * of spam quarantining sets the fate of this mail. -- 2007-03-07/wzk
	 */

	if (x->rbl.blacklisted != 0  &&  x->config->rbl.mode == RBL_QUARANTINE) {
		virulent = IS_SPAM;
		spamscore = x->config->spamd.treshhold;
		}

	/*
	 * Scan for spam, copy&paste from pop3.proxy -- 2007-01-07/wzk
	 */

	else if (x->config->spamd.scan == 0)
		/* we don't have to scan for spam */ ;
	else if (x->size >= SPAM_MAXSIZE) {
		if (debug != 0)
			printerror(0, "+INFO", "e-mail to large for spam scan: size= %ld bytes", x->size);
		}
	else {
		int	rc, pid, pfd[2];

		writestatfile(x, "SPAMSCAN");
		rc = 0;

		if (pipe(pfd) != 0)
			printerror(1, "-ERR", "can't pipe: error= %s", strerror(errno));
		else if ((pid = fork()) < 0)
			printerror(1, "-ERR", "can't fork spamc: error= %s", strerror(errno));
		else if (pid == 0) {
			int	argc, fd;
			char	*argv[32], line[300];

			close(pfd[0]);
			if ((fd = open(x->spoolfile, O_RDONLY)) < 0  ||  dup2(fd, 0) != 0) {
				printerror(1, "-ERR", "can't open spoolfile: filename= %s, error= %s",
						x->spoolfile, strerror(errno));
				}

			dup2(pfd[1], 1);

			if ((fd = open("/dev/null", O_WRONLY)) < 0  ||  dup2(fd, 2) != 2) {
				printerror(0, "-INFO", "can't close stderr, error= %s",
						strerror(errno));
				close(2);	/* then simply close it */
				}


			copy_string(line, x->config->spamd.cmd, sizeof(line));
			argc = split(line, argv, ' ', 30);
			argv[argc] = NULL;
			execvp(argv[0], argv);

			printerror(1, "-ERR", "can't exec spamc %s: error= %s", argv[0], strerror(errno));
			exit (1);
			}
		else {
			char	line[300];
			bio_t	bio;

			close(pfd[1]);		/* not used */

			memset(&bio, 0, sizeof(bio_t));
			bio.fd = pfd[0];
			if (readline_fd(x, &bio, line, sizeof(line), 30, 1) == NULL) {
				int	s;

				s = wait(&rc);
				printerror(0, "-INFO", "spamc error, pid= %d, rc= %d/%d, error= no reply from spamc",
						pid, s, rc, strerror(errno));
				}
			else {
				int	treshhold;

				copy_string(spamlevel, line, sizeof(spamlevel));
				if (wait(&rc) < 0) {
					printerror(0, "-INFO", "process wait error, error= %s", strerror(errno));
					rc = 0;
					}

				if (debug != 0)
					printerror(0, "+INFO", "isspam= %d, spamlevel= %s", rc, spamlevel);

				p = spamlevel;
				spamscore = strtoul(p, &p, 10);
				if (*p != 0)
					p++;

				treshhold = strtoul(p, &p, 10);
				if (treshhold > 0  &&  spamscore > treshhold)
					virulent = IS_SPAM;
				else if (rc != 0)
					virulent = IS_SPAM;
				}

			close(pfd[0]);
			}
		}


	/*
	 * Quarantine spam if configured and above the level.
	 */

	if (virulent == IS_SPAM  &&  x->config->spamd.quarantine != 0
		  &&  spamscore >= x->config->spamd.treshhold
		  &&  *x->config->spamd.qd.dir != 0) {

		char	line[400];

		snprintf (line, sizeof(line) - 2, "%d/%d", spamscore, x->config->spamd.treshhold);
		quarantine(x, &x->config->spamd.qd, x->spoolfile, IS_SPAM, "spamscore", line);

		rc = 550;
		snprintf (resp, size - 2, "%3d spam detected",
					x->config->spamd.statuscode);

		return (rc);
		}

	/*
	 * Send e-mail to server.
	 */

	writestatfile(x, "FORWARD");

	if ((rc = sfputc(x, "DATA", "", resp, size, NULL)) != 354)
		return (rc);

	if ((fp = fopen(x->spoolfile, "r")) == NULL)
		printerror(1, "-ERR", "can't open spoolfile: filename= %s, error= %s", x->spoolfile, strerror(errno));


	/*
	 * Send some additional headers ...
	 */

	if (x->config->scanmail != 0) {
		snprintf (line, sizeof(line) - 2, "X-Proxy-Scanned: %s/%s %s", program, VERSION, x->clamav.version);
		sfputs(x, 2, "%s", line);
		}

	if (x->config->spamd.scan != 0) {
		snprintf (line, sizeof(line) - 2, "X-Spam-Score: %s, %s",
				(virulent == IS_SPAM)? "yes": "no", spamlevel);
		sfputs(x, 2, "%s", line);
		}

	/*
	 * ... followed by the e-mail.
	 */

	spooldata(x, fp, virulent == IS_SPAM? x->config->spamd.spamtag: "");
	fclose (fp);
	rc = sfputc(x, ".", "", resp, size, NULL);

	return (rc);
}

