
/*

    File: smtp.c
    
    Copyright (C) 1999,2000,2003-2007 by Wolfgang Zekoll <wzk@quietsche-entchen.de>

    This source is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This source is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <errno.h>
#include <stdarg.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <signal.h>
#include <netdb.h>
#include <netinet/in.h>

#include "lib.h"
#include "smtp.h"
#include "acp.h"
#include "ip-lib.h"


	/*
	 * Wait timeout sec/usec for the client to talk to us.  The
	 * function is called before the proxy has send it's greeting
	 * to the client to detect so call "early talkers", spam
	 * engines without real SMTP knowledge.  27MAR06wzk
	 */

int checkearlytalker(smtp_t *x, bio_t *bio, int timeout, char *buffer, int size)
{
	int	rc;
	struct timeval tov;
	fd_set	fdset;


	FD_ZERO(&fdset);
	FD_SET(bio->fd, &fdset);

	/* Timeout values below 10 are seconds, greater than 10
	 * milliseconds.
	 */

	if (timeout < 10) {
		tov.tv_sec  = timeout;
		tov.tv_usec = 0;
		}
	else {
		tov.tv_sec  = timeout / 1000;
		tov.tv_usec = timeout % 1000;
		}

	rc = select(bio->fd + 1, &fdset, (fd_set *) NULL, (fd_set *) NULL, &tov);
	if (rc > 0) {
		/* client send something - too early for us */

		if (buffer != 0  &&  size > 0) {
			read(bio->fd, buffer, size - 2);
			noctrl(buffer);
			}

		return (1);
		}

	return (0);
}


	/*
	 * I/O functions - taken from ftp.proxy
	 */

int getc_fd(smtp_t *x, bio_t *bio, int nooptime)
{
	int	c, withkeepalive, SMTP_SERVERTIMER = 60;
	unsigned long sessiontimeout, now, timeoutval;

	withkeepalive = (bio == &x->sbio)? 0: 1;	/* dirty hack. */
	timeoutval = -1;

	if (bio->here >= bio->len) {
		int	rc, max, bytes;
		struct timeval tov;
		fd_set	available, fdset;

		bio->len = bio->here = 0;

		FD_ZERO(&fdset);
		FD_SET(bio->fd, &fdset);
		max = bio->fd;
		bytes = 0;

		/*
		 * When will the whole session timeout?
		 */

		sessiontimeout = time(NULL) + x->config->timeout;

		while (1) {
			available = fdset;
			tov.tv_sec  = x->config->timeout;
			tov.tv_usec = 0;

			if (x->lasttimer == 0  ||  withkeepalive == 0)
				tov.tv_sec = x->config->timeout;
			else {
				unsigned long t1, t2, tt;

				/*
				 * We have to watch two timeouts: the regular
				 * connection timeout (to the client) and a
				 * timeout when we send the server another NOOP
				 * to make him happy while he's waiting for our
				 * next SMTP command.
				 *
				 * t1 is the time till the overall session
				 * timeout, is negative it's already passed and
				 * we terminate.
				 */

				now = time(NULL);
				t1 = sessiontimeout - now;
				if (t1 < 0) {
					printerror(1, "-ERR", "connection timed out, lastcmd= %s", x->lastcmd);
					}


				/*
				 * tt is the number of seconds from the last
				 * keep-alive NOOP ...
				 */

				tt = now - x->lasttimer;
				if (tt < 0  ||  tt > SMTP_SERVERTIMER)
					tt = 0;


				/*
				 * ... and t2 is the time till the next keep-alive.
				 */

				t2 = SMTP_SERVERTIMER - tt;
				if (t2 < 0) {

					/*
					 * This should not happen.
					 */

					printerror(1, "-ERR", "negative timer: t2= %lu", t2);
					}

				/*
				 * Choose the lower timeout.
				 */

				tov.tv_sec = t1 < t2? t1: t2;
				if (debug >= 3) {
					fprintf (stderr, "timeout= %lu, t1= %lu, t2= %lu, now= %lu, last= %lu",
							tov.tv_sec, t1, t2, now, x->lasttimer);
					}
				}

			if (debug >= 3)
				fprintf (stderr, "select max= %d\n", max);

			timeoutval = tov.tv_sec;
			now = time(NULL);
			rc = select(max + 1, &available, (fd_set *) NULL, (fd_set *) NULL, &tov);
			if (rc < 0) {
				printerror(1, "-ERR", "error on select, error= %s", strerror(errno));
				break;
				}


			if (x->lasttimer > 0  &&  withkeepalive != 0  &&  (time(NULL) - x->lasttimer) >= SMTP_SERVERTIMER) {
				char	resp[1024];

				/*
				 * Handle server keep-alive.
				 */

				if (sfputc(x, "NOOP", "", resp, sizeof(resp), NULL) != 250)
					printerror(1, "-ERR", "protocol error on NOOP, response= %s", resp);

				x->lasttimer = time(NULL);
				x->keepalivecounter++;

				/*
				 * If no input is available continue
				 * with select() waiting, otherwise
				 * pass on.
				 */

				if (rc == 0)
					continue;
				}
			else if (rc == 0) {
				printerror(0, "-INFO", "connection timed out: client= %s, server= %s, lastcmd= %s, waitfor= %s, kac= %d, timeout= %ld/%ld %ld/%ld",
					x->client, x->config->server, x->lastcmd, bio->name, x->keepalivecounter,
					timeoutval, x->lasttimer, now, time(NULL));
				return (-1);
				}


			if (FD_ISSET(bio->fd, &available)) {
				if ((bytes = read(bio->fd, bio->buffer, sizeof(bio->buffer) - 2)) <= 0) {
					if (debug != 0) {
						if (bytes == 0)
							fprintf (stderr, "received zero bytes on fd %d\n", bio->fd);
						else
							fprintf (stderr, "received %d bytes on fd %d, errno= %d, error= %s\n", bytes, bio->fd, errno, strerror(errno));
						}

					return (-1);
					}

				break;
				}
			}

		bio->len  = bytes;
		bio->here = 0;
		}

	if (bio->here >= bio->len)
		return (-1);

	c = (unsigned char) bio->buffer[bio->here++];
	return (c);
}

char *readline_fd(smtp_t *x, bio_t *bio, char *line, int size, int nooptime, int cutblanks)
{
	int	c, k;

	*line = 0;
	size = size - 2;

	c = getc_fd(x, bio, nooptime);
	if (c < 0)
		return (NULL);
	else if (c == 1) {
		strcpy(line, "\001");
		return (line);
		}

	k = 0;
	while (c > 0  &&  c != '\n'  &&  c != 0) {
		if (k < size)
			line[k++] = c;

		c = getc_fd(x, bio, nooptime);
		}

	if (c == '\n'  &&  k > 0  &&  line[k-1] == '\r')
		k--;

	line[k] = 0;
	if (cutblanks != 0)
		noctrl(line);

	return (line);
}


char *cfgets(smtp_t *x, char *line, int size, int nooptime, int debuglevel)
{
	char	*p;

	*line = 0;
	if ((p = readline_fd(x, &x->cbio, line, size, nooptime, 1)) == NULL)
		return (NULL);
	else if (debug != 0  &&  debug >= debuglevel)
		fprintf (stderr, "CLI >>>: %s\n", p);

	return (line);
}

int cfputs(smtp_t *x, char *line)
{
	char	buffer[2048];

	if (debug)
		fprintf (stderr, ">>> CLI: %s\n", line);

	/*
	 * Make some buggy clients happy - send line and
	 * cr+lf in on call (which is one packet) -- 17MAR06wzk
	 */

	snprintf (buffer, sizeof(buffer) - 2, "%s\r\n", line);
	write(x->cbio.out, buffer, strlen(buffer));

	return (0);
}

char *sfgets(smtp_t *x, char *line, int size)
{
	char *p;

	*line = 0;
	if ((p = readline_fd(x, &x->sbio, line, size, 0, 1)) == NULL)
		return (NULL);
	else if (debug != 0)
		fprintf (stderr, "SVR >>>: %s\n", p);

	return (line);
}

int sfputs(smtp_t *x, int debuglevel, char *format, ...)
{
	char	buffer[4096];	/* Puffer auf 4k gesetzt, wegen langen Zeilen */
	va_list	ap;

	va_start(ap, format);
	vsnprintf (buffer, sizeof(buffer) - 10, format, ap);
	va_end(ap);

	if (debug != 0  &&  debug >= debuglevel)
		fprintf (stderr, ">>> SVR: %s\n", buffer);

	strcat(buffer, "\r\n");
	write(x->sbio.out, buffer, strlen(buffer));

	return (0);
}

int sfputc(smtp_t *x, char *command, char *parameter, char *line, int size, char **here)
{
	int	rc;
	char	*p, buffer[1030];

	if (command != NULL  &&  *command != 0) {
		if (parameter != NULL  &&  *parameter != 0)
			snprintf (buffer, sizeof(buffer) - 2, "%s %s", command, skip_ws(parameter));
		else
			copy_string(buffer, command, sizeof(buffer));

		sfputs(x, 1, "%s", buffer);
		snprintf (x->lastcmd, sizeof(x->lastcmd) - 2, ">>> SVR: %s", buffer);
		writestatfile(x, buffer);
		}
	
	if (sfgets(x, line, size) == NULL) {
		if (debug != 0)
			printerror(1, "-ERR", "server disappered in sfputc(), pos #1");

		return (-1);
		}

	rc = strtol(line, &p, 10);
	if (*p == '-') {
		while (1) {
			if (sfgets(x, line, size) == NULL) {
				if (debug != 0)
					printerror(1, "-ERR", "server disappered in sfputc(), pos #2");

				return (-1);
				}

			p = noctrl(line);
			if (*p == ' ')
				/* weiter */ ;
			else if (atoi(line) == rc) {
				if (strlen(line) == 3)
					break;
				else if (strlen(line) > 3  &&  line[3] == ' ')
					break;
				}
			}
		}

	p = skip_ws(p);
	if (here != NULL)
		*here = p;

	return (rc);
}


int reset_connection(smtp_t *x)
{
	x->state      = WAITING;

	x->nrcpt      = 0;
	x->rcpt.here  = 0;
/*	*x->rcpt.list = 0; */

	*x->sender    = 0;
	x->nrcpt      = 0;
/*	*x->jobid     = 0; */
	*x->msgid     = 0;
	x->size       = 0;

	setvar("SENDER", "");
	setvar("ORIGSERVER", "");
	setvar("NRCPT", "");
	setvar("RCPT", "");

	return (0);
}




char *get_emailadr(char *envelope, char *email, int size)
{
	char	*p, *r;

	*email = 0;

	/*
	 * Sonderfall: Es ist moeglich, dass die envelope Adresse `<>'
	 * ist.  In diesem liefern wir sie unveraendert zurueck.  Das
	 * beisst sich nicht mit Adressen der Form `<<>>', da geschachtelte
	 * spitze Klammern nicht unterstuetzt werden.
	 */

	if (strcmp(envelope, "<>") == 0) {
		strcpy(email, "<>");
		return (email);
		}


	p = envelope;
	if (*p != '<')
		return (email);
	
	p++;
	if ((r = strchr(p, '>')) == NULL)
		return (email);
	else if (r[1] != 0)
		return (email);

	get_quoted(&p, '>', email, size);
	return (email);
}

int check_emailadr(char *emailadr)
{
	char	*domain;

	if (strcmp(emailadr, "<>") == 0)
		return (1);
	else if (*emailadr == 0)
		return (0);
/*	else if (*emailadr == '@')
 *		return (0);
 */
	else if ((domain = strchr(emailadr, '@')) == NULL)
		return (0);
/*	else if (domain[1] == '0')
 *		return (0);
 */
	else if (strchr(&domain[1], '@') != NULL)
		return (0);
	else if (strchr(emailadr, '!') != NULL  /* ||  strchr(emailadr, '%') != NULL */)
		return (0);
	else if (strchr(emailadr, ':') != NULL  ||  strchr(emailadr, ',') != NULL)
		return (0);
	
	return (1);
}

int _xx_search_allowlist(char *emailadr, char *list)
{
	char	*p, *domain, pattern[200];

	if (strcmp(emailadr, "<>") == 0)
		return (1);

	domain = strchr(emailadr, '@');
	if (domain == NULL  ||  *domain == 0) {

		/*
		 * Das kann eigentlich nicht passieren, die E-Mail Adresse
		 * wurde schon auf genau ein @-Zeichen getestet.
		 */

		return (0);
		}
	else if (list == NULL  ||  *list == 0) {
		
		/*
		 * Kann eigentlich auch nicht vorkommen.
		 */

		return (0);
		}

	p = list;
	while ((p = skip_ws(p)), *get_quoted(&p, ',', pattern, sizeof(pattern)) != 0) {
		noctrl(pattern);
		if (*pattern == '@'  &&  strpcmp(domain, pattern) == 0)
			return (1);
		else if (strpcmp(emailadr, pattern) == 0)
			return (1);
		}
	
	return (0);
}

int search_allowlist(char *emailadr, string_t *list)
{
	char	*p, *domain, pattern[200];

	if (strcmp(emailadr, "<>") == 0)
		return (1);

	domain = strchr(emailadr, '@');
	if (domain == NULL  ||  *domain == 0) {

		/*
		 * Should never happen, all email addresses arriving
		 * here are already checked for an `@' -- 2007-01-23/wzk
		 */

		printerror(0, "-INFO", "email address without domain: %s", emailadr);
		return (0);
		}
	else if (isempty(list)) {

		/*
		 * Can not happen, checked before the function is
		 * called.  However, an empty list means nothing to
		 * check.
		 */

		return (0);
		}

	p = list->string;
	while ((p = skip_ws(p)), *get_quoted(&p, ' ', pattern, sizeof(pattern)) != 0) {
		noctrl(pattern);
		if (*pattern == '@'  &&  strpcmp(domain, pattern) == 0)
			return (1);
		else if (strpcmp(emailadr, pattern) == 0)
			return (1);
		}
	
	return (0);
}


int addrcpt(smtp_t *x, char *rcpt)
{
	int	len;
	char	adr[200];

	/*
	 * Remove angle brackets from e-mail address ...
	 */

	copy_string(adr, &rcpt[1], sizeof(adr));
	len = strlen(adr);
	if (len > 0)
		adr[len-1] = 0;

	/*
	 * ... and add it to the list.
	 */

	len = strlen(rcpt = adr);
	if ((x->rcpt.here + len + 10) >= x->rcpt.max) {
		x->rcpt.max += 1024 + len;
		x->rcpt.list = reallocate(x->rcpt.list, x->rcpt.max);
		}

	if (x->rcpt.here > 0)
		x->rcpt.list[x->rcpt.here++] = ' ';

	strcpy(&x->rcpt.list[x->rcpt.here], rcpt);
	x->rcpt.here += len;

	return (0);
}


int spooldata(smtp_t *x, FILE *fp, char *tag)
{
	int	lineno, isheader;
	char	*p, line[2048];

	lineno   = 0;
	isheader = 1;
	while (1) {
		if (fp != NULL) {
			if (fgets(line, sizeof(line), fp) == NULL)
				break;

			noctrl(line);
			}
		else {
			if (cfgets(x, line, sizeof(line), x->config->timeout, 2) == NULL)
				printerror(1, "-ERR", "client= %s, error= client terminated while sending data", x->client);

			noctrl(line);
			if (line[0] == '.'  &&  line[1] == 0)
				break;
			}



		/*
		 * Insert Received header line.
		 */

		if (lineno == 0) {
			if (x->config->droppath != 2) {
				char	buffer[1024];

				snprintf (buffer, sizeof(buffer) - 2, "Received: from %s (%s) by %s",
						x->client, x->ipnum, x->hostname);
				sfputs(x, 2, "%s", buffer);
				}
			}


		/*
		 * If we are in the mail header we (1) try to read the
		 * mail's message-id and (2) remove the received-path
		 * from the header if removal is configured.
		 */

		if (isheader == 1) {
			if (*line == 0) {
				isheader = 0;

				if (*x->msgid == 0) {
					unsigned long now;
					char	hn[80], line[300];

					gethostname(hn, sizeof(hn) - 2);
					now = time(NULL);

					snprintf (x->msgid, sizeof(x->msgid) - 2, "%lu.%d.%d@%s", now, getpid(), x->mailcount, hn);
					snprintf (line, sizeof(line) - 2, "Message-ID: <%s>", x->msgid);
					sfputs(x, 2, "%s", line);
					printerror(0, "-INFO", "no message id, assuming <%s>, client= %s, sender= %s", x->msgid, x->client, x->sender);
					}
				}
			else {
				char	*p, word[80];

				p = line;
				get_word(&p, word, sizeof(word));
				strlwr(word);
				
				if (strcmp(word, "message-id:") == 0) {
					if (*x->msgid != 0) {
						printerror(0, "-INFO", "duplicate message id: client= %s, sender= %s, message-id: <%s>, duplicate: %s",
								x->client, x->sender, x->msgid, skip_ws(p));
						continue;
						}
					else {
						if ((p = strchr(p, '<')) != NULL) {
							p++;
							get_quoted(&p, '>', x->msgid, sizeof(x->msgid));
							}
						}
					}
				else if (strcmp(word, "received:") == 0) {
					if (x->config->droppath != 0)
						continue;
					}
				else if (*tag != 0  &&  strcmp(word, "subject:") == 0) {
					char	buffer[2048];

					snprintf (buffer, sizeof(buffer) - 2, "%s %s %s",
							word, tag, skip_ws(p));
					copy_string(line, buffer, sizeof(line));

					copy_string(x->subject, p, sizeof(x->subject));
					}
				}
			}


		/*
		 * Send current line to SMTP server.  Notice: If we are
		 * reading from a file we have to provide an additional
		 * dot `.' if the line starts with a dot.
		 */

		p = line;
		if (fp != NULL) {
			if (*line == '.') {
				char	buffer[2050];

				snprintf (p = buffer, sizeof(buffer) - 2, ".%s", line);
				}
			}

		if (sfputs(x, 2, "%s", p) != 0) {
			printerror(0, "-INFO", "server= %s, error= server terminated while receiving data");
			return (-1);
			}

		x->size += strlen(p) + 2;
		lineno++;
		}
		
	return (0);
}


int proxy_request(config_t *config)
{
	int	rc, isearlytalker;
	char	*p, command[10], word[200], line[2048], avgload[80];
	smtp_t	*x;

	/*
	 * Allocate and initialise storage for the connection.
	 */

	x = allocate(sizeof(smtp_t));
	x->config = config;
	x->started = time(NULL);

	copy_string(x->sbio.name, "server", sizeof(x->sbio.name));
	copy_string(x->cbio.name, "client", sizeof(x->cbio.name));
	x->cbio.fd  = 0;
	x->cbio.out = 1;
	x->mailcount = 0;


	get_interface_info(0, &x->interface);

	getpeerinfo(0, x->ipnum, sizeof(x->ipnum), x->client, sizeof(x->client), 0);
	printerror(0, "", "connected to client: client= %s, interface= %s", x->ipnum, x->interface.ipnum);

	setvar("INTERFACE", x->interface.name);
	setvar("CLIENT", x->ipnum);
	setvar("CLIENTNAME", x->client);

	if (*config->configfile != 0) {
		if (readconfig(config, config->configfile, x->interface.ipnum, x->ipnum) == 0)
			printerror(1, "-ERR", "no configuration for interface: %s", x->interface.ipnum);
		}

	/*
	 * Dynamic debug control -- 2007-06-07/wzk
	 */

	if (*config->debug.controlfile != 0  &&  getfs(config->debug.controlfile, line, sizeof(line)) != NULL) {
		if ((debug = strtoul(line, &p, 10)) < 0)
			debug = 0;

		if (debug > 0) {	/* attach debug file to stderr */
			int	fd;

			if (*(p = skip_ws(p)) == 0)
				p = "/tmp";

			snprintf (config->debug.filename, sizeof(config->debug.filename) - 2,
					"%s/smtp-%08lX-%04X.log", p, time(NULL), getpid());
			if ((fd = open(config->debug.filename, O_CREAT | O_WRONLY, 0600)) < 0  ||  dup2(fd, 2) != 2) {
				printerror(0, "-INFO", "cannot open/attach debug logfile, filename= %s, error= %s",
					config->debug.filename, strerror(errno));
				}
			else {
				printerror(0, "", "debuglevel= %d, filename= %s",
					debug, config->debug.filename);
				}

			alsosyslog = 1;
			}
		}


	if (0  &&  config->server.count == 0)
		printerror(1, "-ERR", "no SMTP server configured");

	x->rbl.blacklisted = 0;
	if (isempty(&x->config->rbl.list) == 0) {

		/*
		 * check against the RBL servers if we have some - 2007-01-16/wzk
		 */

		x->rbl.blacklisted = checkrbl(x, x->ipnum, line, sizeof(line));
		if (x->rbl.blacklisted != 0) {
			snprintf (x->rbl.resp, sizeof(x->rbl.resp) - 2, "451 rbl match, %s", line);
			if (x->config->rbl.mode == RBL_REJECT) {
				cfputs(x, x->rbl.resp);		/* was sfputs() -- 2007-06-07/wzk */
				goto end;
				}
			}
		}
	
	if (*config->clientdir != 0) {
		char	logfile[200];
		unsigned long now;
		struct stat sbuf;

		now = time(NULL);
		snprintf (logfile, sizeof(logfile) - 2, "%s/%s", config->clientdir, x->ipnum);
		if (stat(logfile, &sbuf) != 0  ||  (now - sbuf.st_mtime) >= config->accepttime) {
			cfputs(x, "421 service unavailable - authenticate with POP3 first");
			printerror(1, "-ERR", "client not permitted: %s", x->ipnum);
/*			goto end; */
			}
		}


	isearlytalker = 0;
	if (config->earlytalker.timeout > 0) {
		char	rcvd[200];

		*rcvd = 0;
		if (checkearlytalker(x, &x->cbio, config->earlytalker.timeout,
				config->earlytalker.delay != 0? NULL: rcvd,
				sizeof(rcvd) - 2) != 0) {

			if (config->earlytalker.rc < 400)
				config->earlytalker.rc = 421;

			if (config->earlytalker.delay == 0) {

				/*
				 * Terminate now, otherwise wait for the MAIL FROM
				 * to get the sender address for logging purposes.
				 */

				snprintf (line, sizeof(line) - 2, "%d not accepted", config->earlytalker.rc);
				cfputs(x, line);
				printerror(EXIT_ZEROSTATUS, "-ERR", "client error: error= early talker, client= %s, status= %s, received= %s",
								x->ipnum, line, rcvd);
				}

			isearlytalker = 1;
			}
		}

	
	*avgload = 0;
	if (*config->loadavg != 0) {
		if ((rc = checkload(config, config->loadavg, avgload, sizeof(avgload))) != 0) {
			if (rc == 421) {
				cfputs(x, "421 load too high - try again");
				printerror(EXIT_ZEROSTATUS, "-ERR", "average load to high, current= %s", avgload);
				}
			else if (rc != 0) {
				cfputs(x, "500 server error");
				printerror(1, "-ERR", "error on checking load average, rc= %d", rc);
				}
			}
		}

	if (config->server.count > 0) {
		unsigned int port;
		int	fd, start, k;
		char	*srcip, server[200];

		/*
		 * round-robin server mode -- 2007-03-14/wzk
		 */

		start = 0;
		if (x->config->server.count > 0  &&  x->config->server.mode == SM_ROUNDROBIN) {

			/*
			 * We have [0 ... count-1] upstream servers.  Below we cycle
			 * trough them all, we just have to define where to start.  If
			 * the laststartfile is set, we try to read the value from it
			 * to get the current first server.
			 *
			 * Of course, the implementation leaves room for race
			 * conditions.
			 */

			start = -1;
			if (*(p = x->config->server.laststartfile) != 0) {
				FILE	*fp;

				/* Read previous start value. */

				if ((fp = fopen(p, "r")) != NULL) {
					fgets(line, sizeof(line), fp);
					fclose (fp);

					start = atoi(line) + 1;
					if (start >= x->config->server.count)
						start = 0;
					}
				}


			/* If we don't have a value, provide a "reasonable" default */
			if (start < 0)
				start = (getpid() % x->config->server.count);


			if (*(p = x->config->server.laststartfile) != 0) {
				FILE	*fp;

				/* Update the laststartfile. */

				if ((fp = fopen(p, "w")) == NULL) {
					printerror(0, "-INFO", "can't open file: filename= %s, error= %s",
							p, strerror(errno));
					}
				else {
					fprintf (fp, "%d\n", start);
					fclose (fp);
					}
				}
			}

		k = start;
		while (1) {
			copy_string(server, config->server.server[k++], sizeof(server));
			if ((srcip = strchr(server, '/')) != NULL)
				*srcip++ = 0;
			else
				srcip = config->srcip;

			port = get_port(server, 25);
			if ((fd = openip(server, port, srcip, 0)) >= 0)
				break;
			else
				printerror(0, "-INFO", "can't connect to server: %s:%u", server, port);

			if (k > config->server.count)
				k = 0;

			if (k == start)
				break;
			}

		if (fd < 0) {
			printf ("451 Service unavailable\r\n");
			printerror(1, "-ERR", "can't connect to server: %s", config->server.server[start]);
			exit (1);
			}

		x->sbio.fd = x->sbio.out = fd;
		get_client_info(x->sbio.fd, &x->s, 0);
		copy_string(x->servername, server, sizeof(x->servername));
		}
	else if (config->argv != NULL) {
		int	pid, pin[2], pout[2];

		if (pipe(pin) != 0  ||  pipe(pout) != 0) {
			printf ("451 Service unavailable\r\n");
			printerror(1, "-ERR", "proxy error: can't pipe, error= %s", strerror(errno));
			}
		else if ((pid = fork()) < 0) {
			printf ("451 Service unavailable\r\n");
			printerror(1, "-ERR", "proxy error: can't fork, error= %s", strerror(errno));
			}
		else if (pid == 0) {
			dup2(pin[1], 1);
			close (pin[0]);
			
			dup2(pout[0], 0);
			close(pout[1]);
			
			close (2);
			execvp(config->argv[0], config->argv);

			printf ("451 Service unavailable\r\n");
			printerror(1, "-ERR", "proxy error: can't exec, path= %s, error= %s",
						config->argv[0], strerror(errno));
			}
		else {	
			x->sbio.fd = pin[0];
			close(pin[1]);
			
			x->sbio.out = pout[1];
			close(pout[0]);

			copy_string(x->servername, config->argv[0], sizeof(x->servername));
			}
		}
	else {
		printf ("451 Service unavailable\r\n");
		printerror(1, "-ERR", "no proxy server specified");
		}


	printerror(0, "", "connected to server: %s%s%s", x->servername,
				*avgload != 0? ", avgload= ": "", avgload);



	/* Hostnamen holen
	 */

	gethostname(word, sizeof(word));
	getdomainname(line, sizeof(line));
	snprintf (x->hostname, sizeof(x->hostname) - 2, word, line);
	

	/* Greeting Message vom Sendmail Server lesen, und an
	 * Client schicken.
	 */

	sfgets(x, line, sizeof(line));
	while (line[3] != ' ') {
		if (sfgets(x, line, sizeof(line)) == NULL) {
			printerror(1, "-ERR", "lost server while reading server greeting");
			exit (1);
			}
		}
		
	cfputs(x, line);


	/* Wir stellen uns beim lokalen Sendmail Server vor. Die
	 * EHLO-replys werden 'verschluckt' und durch einen eigenen
	 * ersetzt.
	 */

/*	sfputs(x, "%s", "EHLO localhost");
 *	while (sfgets(x, line, sizeof(line)) != NULL) {
 *		if (line[3] == ' ')
 *			break;
 *		}
 *
 *	rc = atol(line);
 */

	rc = sfputc(x, "EHLO", "localhost", line, sizeof(line), NULL);
	if (rc != 250) {
		printerror(0, "-INFO", "server HELO status is not 250, resp= %s", line);
		cfputs(x, "421 service unavailable");

		exit (1);
		}
					



	/*
	 *  **  S M T P   M A I N L O O P
	 */

	x->state = WAITING;
	while (1) {
		rc = 0;		/* Server response code loeschen */


		/*
		 * Naechstes Kommando vom Client holen.
		 */

		if (cfgets(x, line, sizeof(line), 0, 1) == NULL) {
			printerror(0, "-INFO", "client closed connection, client= %s", x->client);
			break;
			}


		/* Kommando isolieren
		 */

		p = noctrl(line);
		get_word(&p, command, sizeof(command));
		strupr(command);
		p = skip_ws(p);


		copy_string(x->lastcmd, command, sizeof(x->lastcmd));

		/* QUIT ist immer moeglich.
		 */

		if (strcmp(command, "QUIT") == 0) {
			rc = sfputc(x, "QUIT", "", line, sizeof(line), NULL);
			cfputs(x, line);

			x->state = SEND_QUIT;
			break;
			}


		/* HELP
		 */

		else if (strcmp(command, "HELP") == 0) {
			cfputs(x, "503 no help available");
			}


		/* NOOP 
		 */

		else if (strcmp(command, "NOOP") == 0) {
			rc = sfputc(x, "NOOP", "", line, sizeof(line), NULL);
			cfputs(x, line);
			}


		/* RSET
		 */

		else if (strcmp(command, "RSET") == 0) {
			rc = sfputc(x, "RSET", "", line, sizeof(line), NULL);
			cfputs(x, line);

			reset_connection(x);
			printerror(0, "", "got RSET command, client= %s", x->client);
			}


		/* ETRN
		 */

		else if (strcmp(command, "ETRN") == 0) {
			if (x->config->etrn == 0) {
				cfputs(x, "500 unrecognized command");
				printerror(0, "-INFO", "ETRN request rejected: client= %s", x->client);
				}
			else {
				if (*get_word(&p, word, sizeof(word)) == 0)
					cfputs(x, "500 ETRN needs parameter");
				else {
					rc = sfputc(x, "ETRN", word, line, sizeof(line), NULL);
					cfputs(x, line);

					if (rc != 250) {
						printerror(0, "-INFO", "ETRN rejected by server, client= %s", x->client);
						}
					}
				}
			}


		/* HELO und EHLO sind nun zweimal verfuegbar, und nicht
		 * wie bisher nur einmal -- 28OCT04wzk
		 */

		else if (strcmp(command, "HELO") == 0  ||  strcmp(command, "EHLO") == 0) {
/*			if (x->helloseen != 0)
 *				cfputs(x, "503 duplicate HELO/EHLO");
 */
			if (x->helloseen >= 2)
				cfputs(x, "503 multiple HELO/EHLO");
			else if (*get_word(&p, word, sizeof(word)) == 0) {
				snprintf (line, sizeof(line) - 2, "501 %s requires domain name", command);
				cfputs(x, line);
				}
			else {
				if (strcmp(command, "HELO") == 0) {
					snprintf (line, sizeof(line) - 2, "250 SMTP server v%s ready - %s [%s]", VERSION, x->client, x->ipnum);
					cfputs(x, line);
					}
				else {
					snprintf (line, sizeof(line) - 2, "250-SMTP server v%s ready - %s [%s]", VERSION, x->client, x->ipnum);
					cfputs(x, line);
					
					cfputs(x, "250-8BITMIME");

					if (x->config->etrn != 0)
						cfputs(x, "250-ETRN");

					cfputs(x, "250 HELP");
					}
					
				x->helloseen++;
				}
			}


		/* MAIL, SEND, SOML, SAML
		 *
		 * Laut RFC 821 kann das MAIL Kommando jederzeit abgesetzt
		 * werden, es macht dabei einen impliziten SMTP-Reset.  Der
		 * real existierende Sendmail will davon aber nichts wissen.
		 */

		else if (strcmp(command, "MAIL") == 0  || strcmp(command, "SEND") == 0  ||
		    strcmp(command, "SOML") == 0  || strcmp(command, "SAML") == 0) {
			
			get_quoted(&p, ':', word, sizeof(word));
			if (strcasecmp(word, "FROM") != 0)
				cfputs(x, "500 syntax error");
			else if (*x->sender != 0)
				cfputs(x, "503 sender already specified");
			else {
				int	allowed;
				char	sender[200], emailadr[200];
				
				p = skip_ws(p);
/* printerror(0, "+DEBUG:", "mailfrom= %s", p); */
				get_word(&p, sender, sizeof(sender));
				strlwr(sender);

				/*
				 * If the client is an early talker and we did not
				 * already disconnect, we do it now, logging the
				 * sender's address.  29MAR06wzk
				 */

				if (isearlytalker == 1) {
					snprintf (line, sizeof(line) - 2, "%d not accepted", config->earlytalker.rc);
					cfputs(x, line);
					printerror(EXIT_ZEROSTATUS, "-ERR", "client error: error= early talker, client= %s, sender= %s, status= %s",
									x->ipnum, sender, line);
					}


				/*
				 * Wir machen ein paar grundsaetzliche Tests mit
				 * der Absenderadresse:
				 *
				 *  - Ist die Adresse von spitzen Klammern
				 *    umgeben?
				 * ...
				 */

				allowed = 1;
				get_emailadr(sender, emailadr, sizeof(emailadr));
				if (*emailadr == 0)
					allowed = 0;

				/*
				 * ...
				 *  - Enthaelt die Adresse mindestens ein @-Zeichen?
				 *  - Enthaelt die Adresse genau ein @-Zeichen?
				 *  - Ist in der Adresse kein !- und kein %-Zeichen
				 *    enthalten.
				 * ...
				 */

				else if (check_emailadr(emailadr) == 0)
					allowed = 0;

				/*
				 * ...
				 *  - Schliesslich wird ggf. noch getestet,
				 *    ob die Absenderadresse auch auf der
				 *    allow-Liste steht.
				 *
				 * Mit den Empfaengeradressen werden die gleichen Tests
				 * durchgefuehrt.
				 */

				else if (isempty(&x->config->senderlist) != 0)
					allowed = 1;
				else 
					allowed = search_allowlist(sender, &x->config->senderlist);

				if (allowed == 0) {
					char	line[300];

					snprintf (line, sizeof(line) - 2, "550 not allowed: %s", sender);
					cfputs(x, line);
					printerror(0, "-INFO", "sender not allowed: sender= %s, client= %s", sender, x->client);
					}
				else {
					char	smtpcmd[20];

					/*
					 * Sender is accepted, rewrite sender domain if required.
					 */

					copy_string(x->origsender, sender, sizeof(x->origsender));
					if (*x->config->senderdomain != 0) {
						char	*p, newadr[300];	/* Resized from 80 to 200 -- 23NOV06wzk */

						if ((p = strchr(sender, '@')) != NULL)
							*p = 0;

						p = sender;
						if (*p == '<')		/* Sollte eigentlich immer der Fall sein */
							p++;

						snprintf (newadr, sizeof(newadr) - 2, "<%s@%s>", p, x->config->senderdomain);
						copy_string(sender, newadr, sizeof(sender));
						}

					snprintf (smtpcmd, sizeof(smtpcmd) - 2, "%s FROM:", command);
					rc = sfputc(x, smtpcmd, sender, line, sizeof(line), NULL);
					cfputs(x, line);

					if (rc == 250) {
						copy_string(x->sender, sender, sizeof(sender));
						x->state = MAIL_SEEN;
						printerror(0, "", "sender: %s", x->sender);
						setvar("SENDER", x->sender);
						}
					else {
						printerror(0, "-INFO", "sender rejected: adr= %s, client= %s, resp= %s", sender, x->client, line);
						}
					}
				}
			}


		/* RCPT
		 */

		else if (strcmp(command, "RCPT") == 0) {
			get_quoted(&p, ':', word, sizeof(word));
			if (strcasecmp(word, "TO") != 0)
				cfputs(x, "500 syntax error");
			else if (x->state != MAIL_SEEN  &&  x->state != RCPT_SEEN)
				cfputs(x, "503 specify sender first");
			else {
				int	allowed;
				char	rcpt[200], emailadr[200];
				
				p = skip_ws(p);
				get_word(&p, rcpt, sizeof(rcpt));
				strlwr(rcpt);

				get_emailadr(rcpt, emailadr, sizeof(emailadr));
				if (*emailadr == 0)
					allowed = 0;
				else if (check_emailadr(emailadr) == 0)
					allowed = 0;
				else if (isempty(&x->config->rcpt) != 0)
					allowed = 1;
				else 
					allowed = search_allowlist(emailadr, &x->config->rcpt);


				if (allowed == 0) {
					char	line[300];

					snprintf (line, sizeof(line) - 2, "550 no such user: %s", rcpt);
					cfputs(x, line);
					printerror(0, "-INFO", "recipient not allowed: rcpt= %s, client= %s", rcpt, x->client);
					}
				else {
					rc = sfputc(x, "RCPT TO:", rcpt, line, sizeof(line), NULL);
					cfputs(x, line);

					if (rc == 250  ||  rc == 251) {
						addrcpt(x, rcpt);
						x->nrcpt++;
						x->state = RCPT_SEEN;

						printerror(0, "+INFO", "rcpt: %s", rcpt);
						}
					else {
						printerror(0, "-INFO", "recipient rejected: rcpt= %s, client= %s, resp= %s", rcpt, x->client, line);
						}
					}
				}
			}


		/* DATA
		 */

		else if (strcmp(command, "DATA") == 0) {

			/*
			 * Check delayed blacklisting, reject on
			 * match -- 2007-03-07/wzk
			 */

			if (x->rbl.blacklisted != 0  &&  x->config->rbl.mode == RBL_DELAYED) {
				printerror(0, "-MAIL", "rbl-match rejected on MAIL: rc= %d, server= %s, client= %s, sender= %s, nrcpt= %d, size= %ld, message-id= <%s>, error= %s",
							atoi(x->rbl.resp), x->servername, x->client, x->sender, x->nrcpt, x->size,
							x->msgid, x->rbl.resp);
				cfputs(x, x->rbl.resp);		/* was sfputs() -- 2007-06-07/wzk */
				goto end;
				}

			
			x->mailcount++;
			setvar("RCPT", x->rcpt.list);
			setnumvar("NRCPT", x->nrcpt);

			if (x->state != RCPT_SEEN)
				cfputs(x, "503 specify recipients first");
			else {
				if (*x->config->acp != 0) {
					run_acp(x, line, sizeof(line));

					rc = atoi(line);
					if (rc == 250  ||  rc == 354)
						/* ok for us, continue processing */ ;
					else if (rc >= 400  &&  rc < 600) {
						printerror(0, "-INFO", "acp status: line= %s", line);
						cfputs(x, line);

						sfputc(x, "RSET", "", line, sizeof(line), NULL);
						reset_connection(x);
						if (rc == 421)
							break;

						continue;
						}
					else {
						printerror(0, "-INFO", "acp status error: line= %s", line);
						cfputs(x, "421 server error");
						break;
						}
					}

				if (x->config->scanmail != 0  ||  x->config->spamd.scan != 0) {
					rc = doscanmail(x, line, sizeof(line));
					cfputs(x, line);
					}
				else {
					rc = sfputc(x, "DATA", "", line, sizeof(line), NULL);
					cfputs(x, line);

					if (rc == 354) {
						spooldata(x, NULL, "");
						rc = sfputc(x, ".", "", line, sizeof(line), NULL);
						cfputs(x, line);
/*						strcpy(x->jobid, "-"); */
						}
					}

				if (rc >= 300) {
					printerror(0, "-MAIL", "rc= %d, server= %s, client= %s, sender= %s, nrcpt= %d, size= %ld, message-id= <%s>, error= %s",
								rc, x->servername, x->client, x->sender, x->nrcpt, x->size,
								x->msgid, line);
					}
				else {
					printerror(0, "+MAIL", "rc= %d, server= %s, client= %s, sender= %s, nrcpt= %d, size= %ld, message-id= <%s>, resp= %s",
								rc, x->servername, x->client, x->sender, x->nrcpt, x->size,
								x->msgid, line);
					}

				if (rc != 250) {
					if (sfputc(x, "RSET", "", line, sizeof(line), NULL) != 250)
						printerror(1, "-ERR", "unexpected response on RSET, response= %s", line);
					}

				reset_connection(x);
				x->state = WAITING;
				}
			}


		/* 
		 * Other commands are not supported.
		 */

		else {
			static int unknowncommands = 0;

			cfputs(x, "500 unrecognized command");
			printerror(0, "-INFO", "received unknown command: cmd= %s", command);

			/*
			 * If the number of unknown command reaches a certain
			 * limit we terminate the connection.  20MAR06wzk
			 */

			if (*command == 0  ||  strlen(command) != 4)
				unknowncommands = UNKNOWNCOMMAND_LIMIT;
			else
				unknowncommands++;

			if (unknowncommands >= UNKNOWNCOMMAND_LIMIT) {
				printerror(0, "-INFO", "too much unknown commands");

				/*
				 * Keep the previous mail (if we have any) for further inspection.
				 */

				if (*x->spoolfile == 0)
					printerror (0, "+INFO", "no mail or no spoolfile to save");
				else {
					char	*p, newname[320];

					copy_string(newname, x->spoolfile, sizeof(newname) - 10);
					if ((p = strrchr(newname, '.')) == NULL  ||  strchr(p, '/') != NULL)
						strcat(newname, ".errmail");
					else
						strcpy(p, ".errmail");

					if (rename(x->spoolfile, newname) != 0) {
						if ((p = strrchr(x->spoolfile, '/')) != NULL)
							p++;
						else
							p = x->spoolfile;

						printerror(0, "-INFO", "can't rename spoolfile: %s, error= %s", x->spoolfile, strerror(errno));
						}
					else {
						if ((p = strrchr(newname, '/')) != NULL)
							p++;
						else
							p = newname;

						printerror(0, "+INFO", "previous mail saved, filename= %s", newname);
						}

					*x->spoolfile = 0;	/* Next mail creates new spoolfile. */
					}

				break;
				}
			}


		/*
		 * Final status code checking - this assumes that the `rc'
		 * variable holds the current server status response code.
		 */

		if (rc == 421) {
			printerror(0, "-INFO", "sendmail returned 421, state= %d, command= %s", x->state, command);
			break;
			}
		else if (rc == -1) {
			printerror(0, "", "terminating (sendmail terminated)"); 
			x->state = NO_SENDMAIL;

			break;
			}
		}

	if (x->state != SEND_QUIT  &&  x->state != NO_SENDMAIL) {
		rc = sfputc(x, "QUIT", "", line, sizeof(line), NULL);
		}

end:
	printerror(0, "+OK", "client %s disconnecting, %d mails", x->client, x->mailcount); 
	return (0);
}

